	;--- Sample RAM implementation of a MSX-UNAPI specification
	;    By Konami Man, 1-2010
	;
	;    This code implements a sample mathematical specification, "SIMPLE_MATH"
	;	.db
	;    which has just two functions:
	;       Function 1: Returns HL = L + E
	;       Function 2: Returns HL = L * E
	;    The code installs on a mapped RAM segment.
	;    The RAM helper must have been previously installed.
	;
	;    To create your own implementation:
	;    1) In the "Constants" section, modify the API version and implementation version
	;       constants, as well as MAX_FN and MAX_IMPFN.
	;    2) In the "Functions code" section, leave FN_INFO unmodified,
	;       and add your own functions.
	;    3) Add an entry for each function in the "Functions addresses table" section.
	;    4) In the "Data" section, modify the specification identifier
	;       and the implementation identifier.
	;
	;    Optional improvements (up to you):
	;    - Install a RAM helper if none is detected.
	;    - Let the user choose the install segment in DOS 1.
	;    - Add code for uninstallation.


;*******************
;***  CONSTANTS  ***
;*******************



;--- System variables and routines

_TERM0		.equ	0x00
_STROUT		.equ	0x09

BDOS		.equ	0x0005
ENASLT		.equ	0x0024
EXTBIO		.equ	0xFFCA
ARG			.equ	0xF847
SEG_CODE	.equ	0x0800


;***************************
;***  INSTALLATION CODE  ***
;***************************

	; org	#100
.area	_CODE

	;--- Show welcome message

	ld		de,#WELCOME_S
	ld		c,#_STROUT
	call	#BDOS

	;--- Locate the RAM helper, terminate with error if not installed

	ld		de,#0x2222
	ld		hl,#0
	ld		a,#0xFF
	call	EXTBIO
	ld		a,h
	or		l
	jr		nz,HELPER_OK

	ld		de,#NOHELPER_S
	ld		c,#_STROUT
	call	BDOS
	ld		c,#_TERM0
	jp		BDOS
HELPER_OK:
	ld		(HELPER_ADD),hl
	ld		(MAPTAB_ADD),bc

	;--- Check if we are already installed.
	;    Do this by searching all the TCP/IP
	;    implementations installed, and comparing
	;    the implementation name of each one with
	;    our implementation name.

	;* Copy the implementation identifier to ARG

	ld	hl,#UNAPI_ID
	ld	de,#ARG
	ld	bc,#UNAPI_ID_END-UNAPI_ID
	ldir

	;* Obtain the number of installed implementations

	ld	de,#0x2222
	xor	a
	ld	b,#0
	call	EXTBIO
	ld	a,b
	or	a
	jr	z,NOT_INST

	;>>> The loop for each installed implementations
	;    starts here, with A=implementation index

IMPL_LOOP:	push	af

	;* Obtain the slot, segment and entry point
	;  for the implementation

	ld	de,#0x2222
	call	EXTBIO
	ld	(ALLOC_SLOT),a
	ld	a,b
	ld	(ALLOC_SEG),a
	ld	(IMPLEM_ENTRY),hl

	;* If the implementation is in page 3
	;  or in ROM, skip it

	ld	a,h
	and	#0b10000000
	jr	nz,NEXT_IMP
	ld	a,b
	cp	#0xFF
	jr	z,NEXT_IMP

	;* Call the routine for obtaining
	;  the implementation information

	ld	a,(ALLOC_SLOT)
	;ld	iyh,a
	.db	0xFD; iy
	ld	h,a

	ld	a,(ALLOC_SEG)
	;ld	iyl,a
	.db	0xFD; iy
	ld	l,a
	ld	ix,(IMPLEM_ENTRY)
	ld	hl,(HELPER_ADD)
	xor	a
	call	CALL_HL	;Returns HL=name address

	;* Compare the name of the implementation
	;  against our own name

	ld	a,(ALLOC_SEG)
	ld	b,a
	ld	de,#APIINFO ; -SEG_CODE_START+SEG_CODE
	ld	ix,(HELPER_ADD)
	inc	ix
	inc	ix
	inc	ix	;Now IX=helper routine to read from segment
NAME_LOOP:	ld	a,(ALLOC_SLOT)
	push	bc
	push	de
	push	hl
	push	ix
	call	CALL_IX
	pop	ix
	pop	hl
	pop	de
	pop	bc
	ld	c,a
	ld	a,(de)
	cp	c
	jr	nz,NEXT_IMP
	or	a
	inc	hl
	inc	de
	jr	nz,NAME_LOOP

	;* The names match: already installed

	ld	de,#ALINST_S
	ld	c,#_STROUT
	call	5
	ld	c,#_TERM0
	jp	5

	;* Names don't match: go to the next implementation

NEXT_IMP:	pop	af
	dec	a
	jr	nz,IMPL_LOOP

	;* No more implementations:
	;  continue installation process

NOT_INST:

	;--- Obtain the mapper support routines table, if available

	xor	a
	ld	de,#0x0402
	call	EXTBIO
	or	a
	jr	nz,ALLOC_DOS2

	;--- DOS 1: Use the last segment on the primary mapper

	ld	hl,(MAPTAB_ADD)
	ld	b,(hl)
	inc	hl
	ld	a,(hl)
	jr	ALLOC_OK

	;--- DOS 2: Allocate a segment using mapper support routines

ALLOC_DOS2:
	ld	a,b
	ld	(PRIM_SLOT),a
	ld	de,#ALL_SEG
	ld	bc,#15*3
	ldir

	ld	a,(PRIM_SLOT)
	or	#0b00100000	;Try primary mapper, then try others
	ld	b,a
	ld	a,#1		;System segment
	call	ALL_SEG
	jr	nc,ALLOC_OK

	ld	de,#NOFREE_S	;Terminate if no free segments available
	ld	c,#_STROUT
	call	5
	ld	c,#_TERM0
	jp	5

ALLOC_OK:
	ld	(ALLOC_SEG),a
	ld	a,b
	ld	(ALLOC_SLOT),a

	;--- Switch segment, copy code, and setup data

	call	GET_P1		;Backup current segment
	ld	(P1_SEG),a

	ld	a,(ALLOC_SLOT)	;Switch slot and segment
	ld	h,#0x40
	call	ENASLT
	ld	a,(ALLOC_SEG)
	call	PUT_P1

	ld	hl,#0x4000	;Clear the segment first
	ld	de,#0x4001
	ld	bc,#0x4000-1
	ld	(hl),#0x0
	ldir

	ld	hl,#SEG_CODE	;Copy the code to the segment
	ld	de,#0x4000
	ld	bc,#0x4000 ; note use file lenth..
	ldir

	ld	hl,(ALLOC_SLOT)	;Setup slot and segment information
	;ld	(MY_SLOT),hl
	;* Now backup and #0xpatc the EXTBIO hook
	;  so that it calls address #4000 of the allocated segment
	;ld	hl,#EXTBIO
	;ld	de,#OLD_EXTBIO
	;ld	bc,#5
	;ldir

	LD	a,#128 ; UNAPIE_INIT, hl=slot , DE=5b jump
	LD	BC,#EXTBIO
	ld	de,#0x2222
	call 0x4000

	;LD	a,#0 ; UNAPI_GET_INFO
	;ld	de,#0x2222
	;call 0x4000
	;LD		D,H
	;LD		E,L
	;ld		c,#_STROUT
	;call	BDOS

	LD	a,#130 ; DRV_INIT
	ld	de,#0x2222
	call 0x4000

	LD	a,#131 ; DRV_GET_INFO
	ld	de,#0x2222
	call 0x4000
	LD		D,H
	LD		E,L
	ld		c,#_STROUT
	call	BDOS

	di
	ld	a,#0xCD	;Code for #"CALL"
	ld	(EXTBIO),a
	ld	hl,(HELPER_ADD)
	ld	bc,#6
	add	hl,bc	;Now HL points to segment call routine
	ld	(EXTBIO+1),hl

	ld	hl,(MAPTAB_ADD)
	ld	a,(ALLOC_SLOT)
	ld	d,a
	ld	e,#0	;Index on mappers table
SRCHMAP:
	ld	a,(hl)
	cp	d
	jr	z,MAPFND
	inc	hl
	inc	hl	;Next table entry
	inc	e
	jr	SRCHMAP
MAPFND:
	ld	a,e	;A = Index of slot on mappers table
	rrca
	rrca
	and	#0b11000000	;Entry point #4010 = index 0
	ld	(EXTBIO+3),a

	ld	a,(ALLOC_SEG)
	ld	(EXTBIO+4),a
	ei

	;--- Restore slot and segment, and terminate

	ld		a,(PRIM_SLOT)
	ld		h,#0x40
	call	ENASLT
	ld		a,(P1_SEG)
	call	PUT_P1

	ld		de,#OK_S
	ld		c,#_STROUT
	call	BDOS

	ld		c,#_TERM0
	jp		BDOS

	;>>> Other auxiliary code

CALL_IX:	jp	(ix)
CALL_HL:	jp	(hl)


;****************************************************
;***  DATA AND STRINGS FOR THE INSTALLATION CODE  ***
;****************************************************

	;--- Variables

PRIM_SLOT:	.db	0	;Primary mapper slot number
P1_SEG:	.db	0		;Segment number for TPA on page 1
ALLOC_SLOT:	.db	0	;Slot for the allocated segment
ALLOC_SEG:	.db	0	;Allocated segment
HELPER_ADD:	.dw	0	;Address of the RAM helper jump table
MAPTAB_ADD:	.dw	0	;Address of the RAM helper mappers table
IMPLEM_ENTRY:	.dw	0	;Entry point for implementations

	;--- DOS 2 mapper support routines

ALL_SEG:	.ds	3
FRE_SEG:	.ds	3
RD_SEG:	.ds	3
WR_SEG:	.ds	3
CAL_SEG:	.ds	3
CALLS:	.ds	3
PUT_PH:	.ds	3
GET_PH:	.ds	3
PUT_P0:	.ds	3
GET_P0:	.ds	3
PUT_P1:	out	(#0xFD),a
	ret
GET_P1:	in	a,(#0xFD)
	ret
PUT_P2:	.ds	3
GET_P2:	.ds	3
PUT_P3:	.ds	3

	;--- Strings

WELCOME_S:
	.str	"UNAPI TCP/IP over MIDI implementation 1.0"
		.db		13,10
	.str	"(c) 2019 by Willem Cazander"
		.db		13,10
	;.db	13,10
	.str	"$"

NOHELPER_S:
	.str	"*** ERROR: No UNAPI RAM helper is installed"
		.db		13,10,"$"

NOMAPPER_S:
	.str	"*** ERROR: No mapped RAM found"
		.db		13,10,"$"

NOFREE_S:
	.str	"*** ERROR: Could not allocate any RAM segment"
		.db		13,10,"$"

OK_S:	.str	"Installed. Have fun!"
		.db		13,10,"$"

ALINST_S:	.str	"*** Already installed."
		.db		13,10,"$"

UNAPI_ID:
		.str	"TCP/IP"
		.db		0
UNAPI_ID_END:

APIINFO:
		.str	"mmidrv-rpc"
		.db		0
APIINFO_END:

.area	_DATA
