
extern crate midir;
//#[macro_use]
extern crate structopt;

use std::thread::sleep;
use std::time::Duration;
use std::io::{stdin, stdout, Write};
use std::error::Error;

use structopt::StructOpt;
use midir::{
	MidiInput,
	MidiOutput,
	MidiInputConnection,
	MidiOutputConnection
};

#[cfg(not(windows))]
use midir::os::unix::VirtualInput;
#[cfg(not(windows))]
use midir::os::unix::VirtualOutput;

const boot_logo: &str = "
Starting:            _             _       
 _ __ ___  _ __ ___ (_) __ _  __ _| |_ ___ 
| '_ ` _ \\| '_ ` _ \\| |/ _` |/ _` | __/ _ \\
| | | | | | | | | | | | (_| | (_| | ||  __/
|_| |_| |_|_| |_| |_|_|\\__, |\\__,_|\\__\\___|
                       |___/               
";

#[derive(StructOpt, Debug)]
#[structopt(name = "mmigate", rename_all = "kebab-case")]
struct Startup {
    /// Enable debug mode.
    #[structopt(short, long)]
    debug: bool,

    /// MIDI client name.
    #[structopt(short, long, default_value = "mmigate")]
    client_name: String,
	
	#[structopt(subcommand)]
    cmd: StartupCommand
}

#[derive(StructOpt, Debug)]
enum StartupCommand {

    /// Proxy internet over virtual MIDI.
    #[structopt(name = "virtual")]
    VirtualProxy {
		/// MIDI client name.
		#[structopt(short, long, default_value = "rpc-tcp-unapi")]
		port_name: String,
	},

	/// List all MIDI devices.
    #[structopt(name = "hw-list")]
    HardwareList {},

	/// Proxy internet over MIDI devices.
    #[structopt(name = "hw-proxy")]
    HardwareProxy {
		/// MIDI-In port
		#[structopt(short = "i", long)]
		midi_in: usize,

		/// MIDI-Out port
		#[structopt(short = "o", long)]
		midi_out: usize,
    },
}

fn main() {
    match run(Startup::from_args()) {
        Ok(_) => (),
        Err(err) => println!("Error: {}", err.description())
    }
}

//const MIDI_CLIENT_NAME: &str = "mmigate";

fn run(startup: Startup) -> Result<(), Box<Error>> {
    println!("{}", boot_logo);
    println!("version: {}", 123);
	if startup.debug {
	    println!("{:?}", startup);
	}
    let midir_in = MidiInput::new(&(startup.client_name.to_string() + "-in"))?;
    let midir_out = MidiOutput::new(&(startup.client_name.to_string() + "-out"))?;

    match startup.cmd {
        StartupCommand::HardwareList {} => {
			//println!("Available MIDI-in ports:");
			for i in 0..midir_in.port_count() {
		        println!("in_{}: {}", i, midir_in.port_name(i)?);
		    }
		    //println!("\nAvailable MIDI-out ports:");
		    for i in 0..midir_out.port_count() {
		        println!("out_{}: {}", i, midir_out.port_name(i)?);
		    }
			return Ok(());
		},
        StartupCommand::VirtualProxy {port_name} => {
			#[cfg(not(windows))] {
                println!("mode: virtual");
                println!("port_names: {}-in,{}-out", port_name, port_name);

				let mut conn_out = midir_out.create_virtual(&(port_name.to_string() + "-out"))?;

				let conn_in = midir_in.create_virtual(&(port_name.to_string() + "-in"), move |stamp, message, _| {
					conn_out.send(message).unwrap_or_else(|_| println!("Error when forwarding message ..."));
					println!("{}: {:?} (len = {})", stamp, message, message.len());
				}, ())?;


				//input.clear();
				let mut input = String::new();
			    stdin().read_line(&mut input)?; // wait for next enter key press
				println!("Closing connections");

				return Ok(());
				//return proxy();
			}
			// 
			#[cfg(windows)] {
				return Error(("Virtual MIDI ports are not supported on Windows."));
			}
		}
        StartupCommand::HardwareProxy {midi_in, midi_out} => {

			// Get an output port (read from console if multiple are available)
			let out_port = match midir_out.port_count() {
				0 => return Err("no output port found".into()),
				1 => {
				    println!("Choosing the only available output port: {}", midir_out.port_name(0).unwrap());
				    0
				},
				_ => {
				    println!("\nAvailable output ports:");
				    for i in 0..midir_out.port_count() {
				        println!("{}: {}", i, midir_out.port_name(i).unwrap());
				    }
				    print!("Please select output port: ");
				    stdout().flush()?;
				    let mut input = String::new();
				    stdin().read_line(&mut input)?;
				    input.trim().parse()?
				}
			};
			
			println!("\nOpening connection");
			let mut conn_out = midir_out.connect(out_port, "midir-test")?;
			println!("Connection open. Listen!");
			{
				// Define a new scope in which the closure `play_note` borrows conn_out, so it can be called easily
				let mut play_note = |note: u8, duration: u64| {
				    const NOTE_ON_MSG: u8 = 0x90;
				    const NOTE_OFF_MSG: u8 = 0x80;
				    const VELOCITY: u8 = 0x64;
				    // We're ignoring errors in here
				    let _ = conn_out.send(&[NOTE_ON_MSG, note, VELOCITY]);
				    sleep(Duration::from_millis(duration * 150));
				    let _ = conn_out.send(&[NOTE_OFF_MSG, note, VELOCITY]);
				};
				
				play_note(66, 4);
				play_note(65, 3);
				play_note(63, 1);
				play_note(61, 6);
				play_note(59, 2);
				play_note(58, 4);
				play_note(56, 4);
				play_note(54, 4);
			}
			sleep(Duration::from_millis(150));
			println!("\nClosing connection");
			// This is optional, the connection would automatically be closed as soon as it goes out of scope
			conn_out.close();
			println!("Connection closed");
			Ok(())
		}
    }
    

}

fn proxy<T>(startup: Startup, conn_in: MidiInputConnection<T>, conn_out: MidiOutputConnection) -> Result<(), Box<Error>> {
	println!("Starting mmigate proxy");

	sleep(Duration::from_millis(150));
    println!("\nClosing connections");
    conn_in.close();
    conn_out.close();
    println!("Connections closed");
    Ok(())
}

